// COPYRIGHT LICENSE:  This information contains sample code provided in source code form.
// You may copy, modify, and distribute these sample programs in any form without payment
// to IBM for the purposes of developing, using, marketing or distributing application
// programs conforming to the application programming interface for the operating platform
// for which the sample code is written. Notwithstanding anything to the contrary,
// IBM PROVIDES THE SAMPLE SOURCE CODE ON AN "AS IS" BASIS AND IBM DISCLAIMS ALL WARRANTIES,
// EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED WARRANTIES OR CONDITIONS
// OF MERCHANTABILITY, SATISFACTORY QUALITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE, AND
// ANY WARRANTY OR CONDITION OF NON-INFRINGEMENT.  IBM SHALL NOT BE LIABLE FOR ANY DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR
// OPERATION OF THE SAMPLE SOURCE CODE.  IBM HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS OR MODIFICATIONS TO THE SAMPLE SOURCE CODE.
//
// (C) Copyright IBM Corp. 2007, 2012.
// All Rights Reserved. Licensed Materials - Property of IBM.

package com.ibm.websphere.batch.samples.tests.bds;

import java.io.RandomAccessFile;
import java.io.IOException;
import java.util.Properties;

import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.BatchDataStream;

/**
 * This is a sample batch data stream (bds) based on a simple line-oriented text file.
 * The underlying file must contain newline characters to delineate logical records. 
 * This bds is an abstract class intended for subclass specialization. This class 
 * supports either read or write access, but not read/write. A subclass needs to 
 * instantiate the superclass for either access mode. No further subclass specialization 
 * is required. This class provides checkpoint support for read and write access as 
 * follows:   
 * 
 * read: 
 * 
 * Checkpoints last line read; restarts at next line.
 * 
 * write: 
 * 
 * Checkpoints last line written; restarts by truncating output file to end of 
 * that line. This means the output mode is strictly append.
 * 
 * Not supported by base class:
 * 
 * 1. output mode that supports random update of existing records
 * 
 * 2. combined r/w output mode
 * 
 */

public abstract class TextFileBatchDataStream implements BatchDataStream {

	private int count= 0;
	
    /**
     * Default constructor is private.
     * Subclasses must use ACCESS_MODE
     * constructor to instantiate 
     * superclass.
     */
    private TextFileBatchDataStream() {}
	
	/**
	 * Holds local name by which consuming batch job step references 
	 * an instance of this batch data stream.
	 */
    protected String _logicalname = null;
    
    /**
     * Holds identifier of the batch job step referencing this
     * batch data stream instance. 
     */
    protected String _stepid = null;
    
    /**
     * Keeps track of where the batch data stream is. In this 
     * case, the position gets increased by the number of bytes 
     * for every line read from or written to the underlying file. 
     */
    protected int _position = 0;
   
    /**
     * Holds name of underlying file. 
     */
    protected String _fn = null;
    
    /**
     * Used to access the underlying file 
     */
    protected transient RandomAccessFile _file = null;
    
    /**
     * Holds the batch data stream input properties
     * for this batch data stream instance. These 
     * properties originate in the batch job definition.
     */
    protected java.util.Properties _props = null;    

    /**
     * Contains the subclass-specified access mode for 
     * this BDS - either read or write. 
     */

    /**
     * Internal switch used to detect and report transition
     * from anonymous bds to bds assigned to job/step.
     */
    protected boolean _printMappingMsg= false;    

    /**
     * Holds access mode indicating how this bds
     * can be accessed, according to ACCESS_MODE
     * enum - currently only read or write access.
     */    
    protected ACCESS_MODE _access_mode; 
    
    /**
     * Defines supported access modes:
     * R= read
     * W= write
     */    
    protected enum ACCESS_MODE {R,W};
    
    /**
     * Required constructor exposed to subclass.
     * @param access_mode specifies access mode for 
     * this BDS, based on ACCESS_MODE enum.
     */
    
    protected TextFileBatchDataStream(ACCESS_MODE access_mode) {
    	_access_mode= access_mode;
    }    

    /**
     * Formats identification string for this bds.
     * @return string for informational messages
     */
    protected String debugString() {
    	if ( _stepid == null ) {
    		_printMappingMsg= true;
    		return "TextFileBatchDataStream["+this+":"+_access_mode+"]";
    	}
    	else {
    		if (_printMappingMsg) {
    			_printMappingMsg= false;
    			System.out.println("TextFileBatchDataStream["+this+":"+_access_mode+"]=["+_stepid+":"+_logicalname+"]");
    		}
    		return "TextFileBatchDataStream["+_stepid+":"+_logicalname+"]";
    	}
    }
    
	/**
	 * Called by Batch Container upon initialization of the job step 
	 * that defines this stream in its job definition. This method is called after the
	 * Batch Container calls setProperties on this same object, so the initialize method
	 * is a good place to get ready to open the underlying data source.  The Batch Container
	 * will call the open method next.    
	 */ 
    public void initialize(String ilogicalname, String ijobstepid) {
        _logicalname = ilogicalname;
        _stepid = ijobstepid;
        
        System.out.println(debugString()+".initialize");
        
        /*
         *  get FILENAME for this BDS from its properties
         */ 
        
        Properties prop = getProperties();        
        if (prop == null) {
            System.out.println(debugString()+" missing properties");
        }
        _fn = prop.getProperty("FILENAME");      
    }
    
    /**
     * Returns string indicating whether this BDS is created
     * for INPUT or OUTPUT mode.
     * @return IO mode string 
     */
    protected String getIOMode() {
    	if ( _access_mode == ACCESS_MODE.R) 
    		return "INPUT";
    	else
    		return "OUTPUT";
    }
       
	/**
	 * Called by the Batch Container to open this BDS. 
	 */
    public void open() throws BatchContainerDataStreamException {

        System.out.println(debugString()+".open["+getIOMode()+"]");

        /*
         * Attempt to open the file underlying this BDS.
         */        
        try {

            if ( _access_mode == ACCESS_MODE.R) {
            	_file = new RandomAccessFile(_fn,"r");
            }
            else {
            	_file = new RandomAccessFile(_fn,"rw");
            }

        } catch (IOException e) {
        	System.out.println(debugString()+": open failure - IO exception on file "+_fn);
        	throw new BatchContainerDataStreamException(e);
        }
    }    
    
    /**
	 * Called by Batch Container at end of job step to close this BDS.        
	 */
    public void close() throws BatchContainerDataStreamException
	{
    	
    	System.out.println(debugString()+".close");
   		try {
   			_file.close();
   		} catch (Exception e) {
   			System.out.println(debugString()+": error closing input stream.");
   			throw new BatchContainerDataStreamException(e);
   		}
    }
    
	/**
	 * Called by the Batch Container to position the BDS 
	 * at the beginning of the underlying data source. For this BDS implementation, that means the 
	 * first line of the underlying file. This method is called once at the beginning of the job
	 * step. This method is not called if the job is restarting after an interruption.  
	 */    
    public void positionAtInitialCheckpoint() {

    	System.out.println(debugString()+".positionAtInitialCheckpoint");
    	this._position = 0;
    }
    
	/**
	 * Called by the Batch Container when the 
	 * job step accessing this BDS reaches a checkpoint.  The Batch Container calls this method
	 * to receive the restart string required by this BDS to reposition itself upon restart of 
	 * the job that was accessing it.  The Batch Container stores this information in its database
	 * until the job is restarted. In the case of this BDS implementation, the checkpoint information
	 * tracks the file position of the last read or write.
	 */    
    public String externalizeCheckpointInformation() {

    	System.out.println(debugString()+".externalizeCheckpointInformation");
        String chkpt = new Integer(this._position).toString();
        
        System.out.println(">>> Checkpoint - "+count++);
        
        return chkpt;
    }    
    
	/**
	 * Called by the Batch Container upon job restart 
	 * to consume this BDS' last checkpoint information.  This method is called only when a job is 
	 * restarted after a planned or unplanned interruption. In the case of this BDS implementation, 
	 * the checkpoint information indicates the file position of the last read or write done immediately 
	 * before the checkpoint was taken.  The Batch Container will call the positionAtCurrentCheckpoint next.   
	 */
    public void internalizeCheckpointInformation(String chkptinfo) {

    	System.out.println(debugString()+".internalizeCheckpointInformation["+chkptinfo+"]");
        this._position = new Integer(chkptinfo).intValue();
    }

	/**
	 * Called by the Batch Container after internalizeCheckpointInformation
	 * to cause the BDS to use the checkpoint information it received from the Batch Container to position 
	 * itself to the location indicated by the checkpoint data.  In the case of this BDS implemenation, 
	 * the checkpoint data is the file position. This method will reposition to that location.  
	 */    
    public void positionAtCurrentCheckpoint() throws BatchContainerDataStreamException {

    	System.out.println(debugString()+".positionAtCurrentCheckpoint");

        try {        
        	if ( _access_mode == ACCESS_MODE.R ) {    
        		_file.seek(this._position);
        	}
        	else {
        		_file.setLength(this._position);
        		_file.seek(this._position);
        	}
        } catch (IOException e) {
        	throw new BatchContainerDataStreamException(e);
        }
    }
    
    /**
     * Called by Batch Container immediately after checkpoint is taken in case
     * BDS needs to take any additional action related to checkpoint boundary.
     * In the case of this BDS implementation, there is nothing to do.
     */
	public void intermediateCheckpoint()
	{
		System.out.println(debugString()+".intermediateCheckpoint");
	}
    
    /**
     * Called by the batch application to access the next
     * record in the input file.
     * 
     * @return - next record from input file
     */
    public String getNextRecord() throws BatchContainerDataStreamException {

    	System.out.println(debugString()+".getNextRecord");
    	
        String input = null;
        try {            
            input = _file.readLine();
            if ( input != null ) { 
                /*
                 * Not EOF, so increase position counter.
                 */
            	this._position+=input.length()+1;
            }
        } catch (IOException e) {
   			System.out.println(debugString()+": error reading input stream.");
            throw new BatchContainerDataStreamException(e);
        }
        return input;
    }
    
    /**
     * Called by batch application to write next output record.
     * @param line to write 
     * @throws BatchContainerDataStreamException
     */
    public void putNextRecord(String r) throws BatchContainerDataStreamException {
    	
    	System.out.println(debugString()+".putNextRecord");
    	
        try {            
            _file.writeBytes(r);
            _file.write('\n');
            this._position+=r.length()+1;
        } catch (IOException e) {
   			System.out.println(debugString()+": error writing output stream.");
            throw new BatchContainerDataStreamException(e);
        }
    }

    /**
     * Returns logical name of this BDS.
     * @return BDS logical name.
     */
    public String getName() {
    	System.out.println(debugString()+".getName");
        return _logicalname;
    }

    /**
     * Sets input properties object on this BDS.
     */
    public void setProperties(Properties properties) {
    	System.out.println(debugString()+".setProperties");
    	_props = properties;
    }
    
    /**
     * Returns input properties object belonging to this BDS.
     * @return BDS input properties. 
     */
    public Properties getProperties() {
    	System.out.println(debugString()+".getProperties");
    	return _props;
    }

}
