// COPYRIGHT LICENSE:  This information contains sample code provided in source code form.
// You may copy, modify, and distribute these sample programs in any form without payment
// to IBM for the purposes of developing, using, marketing or distributing application
// programs conforming to the application programming interface for the operating platform
// for which the sample code is written. Notwithstanding anything to the contrary,
// IBM PROVIDES THE SAMPLE SOURCE CODE ON AN "AS IS" BASIS AND IBM DISCLAIMS ALL WARRANTIES,
// EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED WARRANTIES OR CONDITIONS
// OF MERCHANTABILITY, SATISFACTORY QUALITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE, AND
// ANY WARRANTY OR CONDITION OF NON-INFRINGEMENT.  IBM SHALL NOT BE LIABLE FOR ANY DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR
// OPERATION OF THE SAMPLE SOURCE CODE.  IBM HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
// SUPPORT, UPDATES, ENHANCEMENTS OR MODIFICATIONS TO THE SAMPLE SOURCE CODE.
//
// (C) Copyright IBM Corp. 2007, 2012.
// All Rights Reserved. Licensed Materials - Property of IBM.
package com.ibm.websphere.batch.samples.tests.steps;

import java.util.Properties;
import com.ibm.websphere.batch.BatchConstants;
import com.ibm.websphere.batch.BatchJobStepInterface;
import com.ibm.websphere.batch.BatchDataStreamMgr;
import com.ibm.websphere.batch.BatchContainerDataStreamException;
import com.ibm.websphere.batch.JobStepID;
import com.ibm.websphere.batch.context.JobStepContextMgr; 
import com.ibm.websphere.batch.context.JobStepContext;
import com.ibm.websphere.batch.samples.tests.bds.TestInputBatchDataStream;
import com.ibm.websphere.batch.samples.tests.bds.TestOutputBatchDataStream;


/**
 * This is a sample batch job step that exemplifies the WebSphere
 * batch programming model.  The batch job step is described
 * by a job description and processed by the WebSphere Batch
 * Container.  The Batch Container invokes the following methods
 * on the batch job step: 
 * 
 * 1.  public void setProperties(java.util.Properties properties)
 * 2.  public java.util.Properties getProperties()  
 * 3.  public void createJobStep() 
 * 4.  public int processJobStep() 
 * 5.  public int destroyJobStep()
 * 
 * The business logic is executed by the processJobStep method. 
 * The input/output data for processJobStep is documented by the
 * job step author as logical names that are mapped to concrete
 * batch data streams (BDS) through a job definition.   
 * 
 * The processJobStep method is invoked repeatedly by the 
 * Batch Container to process the next logical input record(s).
 * This 'batch loop' continues until the processJobStep method
 * returns a result value indicating it has finished.  
 * 
 * This job step reads an input BDS and copies it to an output
 * BDS.  The required BDS logical names are:
 * 
 * 1. input
 * 2. output
 * 
 */

public class TestBatchJobStep implements BatchJobStepInterface {

	
	private final static int SUCCESS=0;
	private Properties _props; 
	private TestInputBatchDataStream _testInputBatchDataStream;
	private TestOutputBatchDataStream _testOutputBatchDataStream;

	private Integer time= 0;
	
	private Integer count= 0;
	
	
	/**
	 * Helper method that returns this job step's ID from the 
	 * job step context manager.  This ID is needed to retrieve
	 * this step's batch data streams (BDS) from the batch
	 * data stream manager.
	 * @return JobStepID representing this job step instance.
	 */
	private JobStepID getJobStepID() { 
		JobStepContext ctx= JobStepContextMgr.getContext();
		return ctx.getJobStepID();
		
		/*
		String jobid= _props.getProperty(BatchConstants.JOB_ID);				
		String stepid= _props.getProperty(BatchConstants.STEP_ID);


		return new JobStepID(jobid, stepid);

		*/
		
	}		
	
	/**
	 * Called from processJobStep to assist with testing
	 * checkpoint restart logic. System property 'failOn'
	 * specifies the input file value to detect.  When
	 * detected, job failure is simulated by existing 
	 * the JVM. 
	 * @param data from input file to compare against value of failOn system property
	 */
	private void exitIfTestingCheckpoint(String data) {
		String failOn= System.getProperty("failOn");
		if ( failOn != null && data.equals(failOn) ) {
			System.out.println("TestBatchJobStep: testing checkoint, exiting after reading "+data);
			Runtime.getRuntime().exit(-1);
		}
	}
	
    /**
     * Called by the Batch Container to make the properties 
     * specified in the job definition available to the batch job step. 
     * @param properties - the properties specified in job definition are 
     * available in this Properties object.
     */
    public void setProperties(java.util.Properties properties) {
    	_props=properties; 
    	System.out.println("TestBatchJobStep: setProperties: "+_props);
    	
    	if ( properties != null ) { 
			String stime= properties.getProperty("sleeptime");
			if ( stime != null ) { 
				try { 
					System.out.println("Converting time value "+stime);
					time= Integer.valueOf(stime).intValue() * 1000;
					System.out.println("Time value in millis= "+time);
				}
				catch (Exception e) { 
					e.printStackTrace();
					System.out.println("Blew up big time converting time to int");
				}
			}
		}
    	

    	
    }
    
    /**
     * Returns the properties specified in the job definition for the batch job step. 
     * @return Properties object containing input properties for this job step.
     */
    public java.util.Properties getProperties() {
    	System.out.println("TestBatchJobStep: getProperties");
    	return _props;
    }
        
    /**
     * Called once by the Batch Container before it calls 
     * processJobStep. Any initialization logic, such as 
     * retrieving a BDS property, can be placed in this method.          
     */
    public void createJobStep() { 
    	System.out.println("TestBatchJobStep: createJobStep");
    	
    	/*
    	 * Input properties passed through the job definition can be retrieved 
    	 * through the job step's Properties object. 
    	 */
    	
    	String title= getProperties().getProperty("title");
    	
    	System.out.println("TestBatchJobStep: property title="+title);
    	
        try {    	    
        	
        	/* 
        	 * The bds logical name and the job step id are needed to retrieve each bds.
        	 */
       
            JobStepID id = getJobStepID();
        	_testInputBatchDataStream=  (TestInputBatchDataStream) BatchDataStreamMgr.getBatchDataStream( "input", id.getJobstepid() );
        	
        	_testOutputBatchDataStream=  (TestOutputBatchDataStream) BatchDataStreamMgr.getBatchDataStream( "output", id.getJobstepid() );
        	
        }
        catch (BatchContainerDataStreamException e) {
        	e.printStackTrace();
            System.out.println("TestBatchJobStep: error accessing batch data stream : "+e);
            throw new RuntimeException (e);
        }
    	
    }

    /**
     * <p>
     * Method processJobStep should contain all the business logic for the batch job step. It is called by the Batch Container
     * in a batch loop. The intention of the batch loop model is that each entry into processJobStep is for the purpose of 
     * processing a single input record. The input may be literally a single record or a single instance of whatever collection 
     * of input represents the "logical" record.  The return code of the processJobStep indicates to the Batch Container whether 
     * it should exit the batch loop or not. These return codes are specified in the BatchConstants class. 
     * 
     * For example, if processJobStep returns with return code BatchConstants.STEP_CONTINUE, the Batch Container will not exit 
     * the batch loop and will call processJobStep again in the next iteration of the batch loop; however, if the processJobStep 
     * returns with return code BatchConstants.STEP_COMPLETE, the Batch Container will exit the batch loop and therefore stop calling 
     * the processJobStep method. 
     * <p>
     * The processJobStep method runs under a global transaction managed by the Batch Container. The Checkpoint policy that is applied 
     * to the batch job (assigned via the job definition) determines when the Batch Container should commit the global transaction.
     * Committing the global transaction means the job step has reached a checkpoint.  If the job step is restarted after a planned
     * or unplanned outage, it resumes from the last checkpoint.
     * 
     * It is the responsibility of the batch job step implementer to ensure that any transaction sensitive calls made in the 
     * processJobStep method, eg: a call to another entity EJB, should inherit the global transaction under which processJobStep runs.  
     * <p>
     * @return a return code specified in BatchConstants
     */
    public int processJobStep() {
    	
    	System.out.println("TestBatchJobStep: processJobStep");

    	try { 
    	
    		/* 
    		 * Retrieve next input record.  
    		 */
    		
    		String l= _testInputBatchDataStream.getNextRecord();
    	
    		System.out.println("TestBatchJobStep: read line from myinput: "+l);
    	
    		if ( l == null ) {
    			/*
    			 * No more input records, to indicate to Batch Container to exit batch loop.
    			 */
    			return BatchConstants.STEP_COMPLETE;	    		
    		}
    		else { 
    			/*
    			 * Copy current input line to output batch data stream.   
    			 */
    			_testOutputBatchDataStream.putNextRecord(l);    			

    			
    			if ( time > 0 ) { 
    			
    				System.out.println("Begin sleep - count="+count++);
    				synchronized (this) {
    					try { 
    						this.wait(time);
    					}
    					catch (Exception e) {
    						e.printStackTrace();
    						System.out.println("Caught exception");
    					}
    				}
    				System.out.println("End sleep ");
    			}  			
    			
    			exitIfTestingCheckpoint(l);
    			
    			/*
    			 * Instruct Batch Container to continue batch loop.
    			 */
    			return BatchConstants.STEP_CONTINUE;	
    		}
    	}
    	catch (Exception e) {
    		e.printStackTrace();
    		throw new RuntimeException("TestBatchJobStep: error in processJobStep: ",e);
    	}	    	
    }	    
    
    /**
     * Called once by Batch Container when it has finished processing the job step. Any clean up code can be added here.
     * The return code returned by destroyJobStep will be the return code of this Batch Step. 
     * @return a return code for this Batch Step. 
     */
    public int destroyJobStep() {
    	System.out.println("TestBatchJobStep: destroyJobStep");
    	return SUCCESS;
    }
	
}
